// -*- C++ -*-
// ACL:license
// ----------------------------------------------------------------------
// This software and ancillary information (herein called "SOFTWARE")
// called POOMA (Parallel Object-Oriented Methods and Applications) is
// made available under the terms described here.  The SOFTWARE has been
// approved for release with associated LA-CC Number LA-CC-98-65.
// 
// Unless otherwise indicated, this SOFTWARE has been authored by an
// employee or employees of the University of California, operator of the
// Los Alamos National Laboratory under Contract No. W-7405-ENG-36 with
// the U.S. Department of Energy.  The U.S. Government has rights to use,
// reproduce, and distribute this SOFTWARE. The public may copy, distribute,
// prepare derivative works and publicly display this SOFTWARE without 
// charge, provided that this Notice and any statement of authorship are 
// reproduced on all copies.  Neither the Government nor the University 
// makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this SOFTWARE.
// 
// If SOFTWARE is modified to produce derivative works, such modified
// SOFTWARE should be clearly marked, so as not to confuse it with the
// version available from LANL.
// 
// For more information about POOMA, send e-mail to pooma@acl.lanl.gov,
// or visit the POOMA web page at http://www.acl.lanl.gov/pooma/.
// ----------------------------------------------------------------------
// ACL:license

#ifndef POOMA_TINY_BINARY_VECTOR_OP_H
#define POOMA_TINY_BINARY_VECTOR_OP_H

//-----------------------------------------------------------------------------
// Class: BinaryVectorEngine
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Overview:
// An engine class for representing the sum, product etc between two
// vectors.  This is used as part of evaluating expressions.
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
//
// Full Description:
//
// This specialization of VectorEngine implements an optimization
// for expression evaluation.  Binary operations are done by constructing
// a Vector with a BinaryVectorOp engine tag, then the result of the 
// operation is constructed by "copying out of" that one.  Copying each 
// element out of one of these causes the binary expression to be evaluated 
// for that point.
// 
//-----------------------------------------------------------------------------

//
// The engine tag is just a tag.
//

template<class V1, class V2, class Op>
class BinaryVectorOp
{
};


template<int D, class T, class V1, class V2, class Op>
class VectorEngine<D,T, BinaryVectorOp<V1,V2,Op> >
{

public:
 
  //----------------------------------------------------------------------
  // Typedefs
  //----------------------------------------------------------------------

  // Export the input types.
  enum { dimensions=D };
  typedef T Element_t;
  typedef BinaryVectorOp<V1,V2,Op> EngineTag_t;

  // Return types for accessor functions.
  typedef T ConstElementRef_t;
  typedef T ElementRef_t;

  // Record the type of the current class.
  typedef VectorEngine<D,T, BinaryVectorOp<V1,V2,Op> > This_t;


  //----------------------------------------------------------------------
  // Constructors and Destructor

  // Construct from two vectors and on operator tag.
  VectorEngine(const V1& v1, const V2& v2, Op op)
    : v1_m(v1), v2_m(v2), op_m(op) {}

  // Construct from two vectors and let the op tag contruct itself.
  VectorEngine(const V1& v1, const V2& v2)
    : v1_m(v1), v2_m(v2) {}

  // Copy ctor just copies the references.
  VectorEngine(const This_t& x)
    : v1_m(x.v1_m), v2_m(x.v2_m), op_m(x.op_m) {}

  // Let the engine destroy itself.
  ~VectorEngine() {}


  //----------------------------------------------------------------------
  // Element access

#if !POOMA_NO_TEMPLATE_FRIENDS
  template<int DD,class TT, class EE, int I>
    friend struct VectorEngineElem;

private:
#endif

  const V1& v1_m;
  const V2& v2_m;
  Op op_m;
};

//-----------------------------------------------------------------------------
//
// Specialization of VectorElem for BinaryVectorOp.
// Compile time element lookup.
//
//-----------------------------------------------------------------------------

template<int D, class T, class V1, class V2, class Op, int I>
struct VectorEngineElem<D,T,BinaryVectorOp<V1,V2,Op>, I >
{
  typedef VectorEngine<D,T,BinaryVectorOp<V1,V2,Op> > V;
  typedef typename VectorElem<V1,I>::Element_t T1;
  typedef typename VectorElem<V2,I>::Element_t T2;
  typedef typename BinaryReturn<T1,T2,Op>::Type_t Element_t;
  typedef Element_t ElementRef_t;
  typedef Element_t ConstElementRef_t;
  static Element_t get(const V& x) 
    { 
      return Op()(
        VectorElem<V1,I>::get(x.v1_m),
        VectorElem<V2,I>::get(x.v2_m));
    }
};

#endif

// ACL:rcsinfo
// ----------------------------------------------------------------------
// $RCSfile: BinaryVectorOp.h,v $   $Author: swhaney $
// $Revision: 1.10 $   $Date: 2000/03/07 13:18:12 $
// ----------------------------------------------------------------------
// ACL:rcsinfo
